
/* binary helper app to patch vems.hex with airXfactor.hex
 * or cltfactor.hex data
 * 
 * known limitations:
 * The upper 64kbyte is handled with a hack
 * (vems.hex not changed above 1024 lines)
 * 
 * 
 */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MAX_CHANGEABLE_LINEIDX (1023)

#define ADDRESS_MIN (0x100)
#define HEXLINE_MINLEN (20)
#define FACTOR_PATCH_SIZE (128)
// each element is a hex-line (16 bytes of data)
char * factor_patch[ FACTOR_PATCH_SIZE ];

#define LINEBUFSIZE 128
char linebuf[ LINEBUFSIZE ];
int linecnt = 0;
	
struct stats_t {
	int lines_remained;	
	int lines_replaced;	
};

struct stats_t stats;

/*
 * removed line-end character
 * (TODO: check for empty-string - if you care. not appropriate here)
 */
void chomp( char * str )
{
	int i;
	int len = strlen( str );
	for( i = len-1; i; i--){
		if( str[i] == '\r' || str[i] == '\n' ){
			str[i] = '\0';
		} else {
			break;	
		} 
	} 	
}

int hexchr_val( char hexchr )
{
	if( hexchr >= '0' && hexchr <= '9' ){
		return hexchr - '0';
	}
	
	if( hexchr >= 'A' && hexchr <= 'F' ){
		return hexchr - 'A' + 10;
	}
	
	if( hexchr >= 'a' && hexchr <= 'F' ){
		return hexchr - 'a' + 10;
	}
	
	fprintf( stderr, "ERROR: not a hexa digit: %c\n", hexchr );
	return -1;
}

/* returns address or -1 if bad;
 * the index for a reasonable line
 *  :10035000A4A3A2A1A0A09F9E9D9C9C9B9A999998C2
 * will return 0x0350
 * */
int factor_address(char *line)
{	
	int address = -1;
	
	if( line == NULL ){
		return -1;
	}
	
	if( strlen( line ) >= HEXLINE_MINLEN
	&& line[0] == ':'
	&& line[1] == '1'
	&& line[2] == '0'
	&& line[7] == '0'
	&& line[8] == '0'
	) {
		address = 0x1000 * hexchr_val(line[3])
			+ 0x100 * hexchr_val(line[4])
			+ 0x10 * hexchr_val(line[5])
			+ hexchr_val(line[6]);
	} else {
		fprintf( stderr, "NOTE: hexfile line neglected: %s", line );
	}
	
	return address;
}

/*
 * returns (address-0x100) / 16
 * or -1 if bad (range-violation)
 */ 
int address_to_idx( int address )
{
	int idx;	
	if( address < 0 ){
		return -1;
	}
	
	// address => idx formula (airdenfactor starts at 0x100)
	idx = (address - ADDRESS_MIN) >> 4;
	if( (address & 0xF) != 0){
		fprintf( stderr, "NOTE: hexfile addr (division by 16) at line %d (addr 0x%x)%s\n",
			linecnt, address, linebuf );
			return -1;
	}
	
	// check for range violation
	if( idx < 0	|| idx >= FACTOR_PATCH_SIZE	){
		return -1;
	}	
	return idx;		
}

/** returns address for certain factor,
 * or -1 if err
 */
int factor_idx( char* line ){
	int address = factor_address( line );
    return address_to_idx( address );
}

/* examines a line and stores where appropriate */
void store_factor(char *line)
{
	int idx;
	if( line == NULL ){
		return;
	}
	idx = factor_idx( line );
	if( idx >= 0 ){
		char* linecopy;
		int len = strlen( line );
		linecopy = malloc( len );
		if( linecopy == NULL ){
			fprintf( stderr, "Error: malloc failed\n");
			exit(-2);
		}
		
		if( strcpy( linecopy, line ) != linecopy ){
			fprintf( stderr, "Error: strcpy failed\n");
			exit(-1);
		}
		
		chomp( linecopy );
		factor_patch[ idx ] = linecopy;
		// fprintf( stderr, "stored to %02x :%s", idx, line );
	}
}
 
 /* returns line for certain factor_old,
  * or NULL
  */
char* get_patchline( int idx )
{
	if( idx < 0	|| idx >= FACTOR_PATCH_SIZE	){
		return NULL;
	} 
	return factor_patch[ idx ];
}

/** reads from PATCHFILE (caller must fclose()!!!) 
 * and call store_factor()
 */ 
void read_patch( FILE * PATCHFILE )
{
	char * line;	
	do{
		line = fgets(linebuf, LINEBUFSIZE, PATCHFILE);
        if( line == NULL ){
        		break;
        }    
        store_factor( line );        
	} while( 1 );
}
/*
 * reads STDIN
 * and for each line, prints old-line or patch-line
 */
void hexpatch( char* oldfilename )
{
	char *replacement = NULL;
	char * line;
	FILE *OLDF;
	
	read_patch( stdin );
	OLDF = fopen( oldfilename, "r" );
	
	do{ 	
		int idx;
		char replaced = 0;
		line = fgets(linebuf, LINEBUFSIZE, OLDF);
        if( line == NULL ){        	
        		break;
        }
        linecnt ++;

		if( linecnt <= MAX_CHANGEABLE_LINEIDX ){
        	idx = factor_idx( line );
        	replacement = get_patchline( idx );
			if( replacement != NULL ){
				// chomed in store_factor
	  			printf( "%s\n", replacement );
	  			fprintf( stderr, "replaced (idx=%x) to: %s\n",idx, replacement );
	  			stats.lines_replaced ++;
	  			replaced = 1;
			}
		} 
		
		if( replaced == 0 ){
			chomp(line);
			printf( "%s\n", line ); // old line
			stats.lines_remained ++;
		}
	} while( 1 );
	
	fclose( OLDF );	
	
	fprintf( stderr,  "stats: %d lines replaced (%d remained)\n", stats.lines_replaced, stats.lines_remained);
}

/*
 * Prints usage
 */ 
void usage( char * bin_name )
{
	printf( "usage:\n\t%s oldfilename.hex < patchlines.hex > patched.hex\n", bin_name );
}

/*
 * STDIN: patch 
 * argument1: oldfile
 * STDOUT: patched output
 */ 
int main(int argc, char *argv[])
{
	char * filenameold;
	if( argc != 2 ){
		usage( argv[0] );
		exit(0);
	}
	
	filenameold = argv[1];
	hexpatch( filenameold );
	return 0;
}
